import collections
import math
import typing as tp
import warnings

from satella.coding.transforms.percentile import percentile
from .base import EmbeddedSubmetrics, MetricLevel
from .measurable_mixin import MeasurableMixin
from .registry import register_metric
from ..data import MetricData, MetricDataCollection


@register_metric
class SummaryMetric(EmbeddedSubmetrics, MeasurableMixin):
    """
    A metric that can register some values, sequentially, and then calculate quantiles from it.
    It calculates configurable quantiles over a sliding window of amount of measurements.

    :param last_calls: last calls to handle() to take into account
    :param quantiles: a sequence of quantiles to return in to_metric_data
    :param aggregate_children: whether to sum up children values (if present)
    :param count_calls: whether to count total amount of calls and total time
    """
    __slots__ = ('last_calls', 'calls_queue', 'quantiles', 'aggregate_children',
                 'count_calls', 'tot_calls', 'tot_time')

    CLASS_NAME = 'summary'

    def __init__(self, name, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[MetricLevel] = None,
                 internal: bool = False,
                 last_calls: int = 100, quantiles: tp.Sequence[float] = (0.5, 0.95),
                 aggregate_children: bool = True,
                 count_calls: bool = True, *args,
                 **kwargs):
        super().__init__(name, root_metric, metric_level, *args, internal=internal,
                         last_calls=last_calls, quantiles=quantiles,
                         aggregate_children=aggregate_children, count_calls=count_calls,
                         **kwargs)
        self.last_calls = last_calls  # type: int
        self.calls_queue = collections.deque()  # type: tp.List[float]
        self.quantiles = quantiles  # type: tp.List[float]
        self.aggregate_children = aggregate_children  # type: bool
        self.count_calls = count_calls  # type: bool
        self.tot_calls = 0  # type: int
        self.tot_time = 0  # type: float

    def _handle(self, time_taken: float, **labels) -> None:
        if self.count_calls:
            self.tot_calls += 1
            self.tot_time += time_taken

        if labels or self.embedded_submetrics_enabled:
            return super()._handle(time_taken, **labels)

        if len(self.calls_queue) == self.last_calls:
            self.calls_queue.pop()

        self.calls_queue.appendleft(time_taken)

    def to_metric_data(self) -> MetricDataCollection:
        k = self._to_metric_data()
        if self.count_calls:
            k += MetricData(self.name + '.count', self.tot_calls, self.labels, self.get_timestamp(),
                            self.internal)
            k += MetricData(self.name + '.sum', self.tot_time, self.labels, self.get_timestamp(),
                            self.internal)
        return k

    def _to_metric_data(self) -> MetricDataCollection:
        if self.embedded_submetrics_enabled:
            k = super().to_metric_data()
            if self.aggregate_children:
                total_calls = []
                for child in self.children:
                    total_calls.extend(child.calls_queue)
                total_calls.sort()

                q = self.calculate_quantiles(total_calls)
                q.postfix_with('total')
                k += q

            if self.count_calls:
                k += MetricData(self.name + '.count', self.tot_calls, self.labels,
                                self.get_timestamp(), self.internal)
                k += MetricData(self.name + '.sum', self.tot_time, self.labels,
                                self.get_timestamp(),
                                self.internal)

            return k
        else:
            return self.calculate_quantiles(self.calls_queue)

    def calculate_quantiles(self, calls_queue) -> MetricDataCollection:
        output = MetricDataCollection()
        sorted_calls = sorted(calls_queue)
        for p_val in self.quantiles:
            if not sorted_calls:
                output += MetricData(self.name, 0.0, {'quantile': p_val, **self.labels},
                                     self.get_timestamp(), self.internal)
            else:
                output += MetricData(self.name, percentile(sorted_calls, p_val),
                                     {'quantile': p_val, **self.labels}, self.get_timestamp(),
                                     self.internal)
        return output


@register_metric
class QuantileMetric(SummaryMetric):
    CLASS_NAME = 'quantile'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn('quantile is deprecated and will be removed in Satella 3.0; '
                      'use summary instead', DeprecationWarning)



import typing as tp
from concurrent.futures import Executor

from satella.coding.concurrent.futures.futures import wrap_if


def call_in_future(executor: Executor, function: tp.Callable,
                   *args, **kwargs) -> tp.Callable[[], 'Future']:
    """
    Return a callable, whose calling will schedule function to be executed on a target Executor.

    The returned function will accept any number of arguments and keyword arguments, but will simply
    ignore them.

    :param executor: executor to run at
    :param function: callable to schedule
    :param args: arguments to provide to the callable
    :param kwargs: keyword arguments to provide to the callable
    :return: a callable, calling which will schedule function to run at executor. Calling this callable
        will return the Future for that function
    """

    def inner(*my_args, **my_kwargs):
        return wrap_if(executor.submit(function, *args, **kwargs))

    return inner



import typing as tp

import collections
import threading

from satella.coding.recast_exceptions import rethrow_as
from satella.coding.concurrent.thread import Condition
from satella.coding.typing import T
from satella.exceptions import WouldWaitMore, Empty
from satella.time import measure


class PeekableQueue(tp.Generic[T]):
    """
    A thread-safe FIFO queue that supports peek()ing for elements.
    """
    __slots__ = ('queue', 'lock', 'inserted_condition')

    def __init__(self):
        super().__init__()
        self.queue = collections.deque()
        self.lock = threading.Lock()
        self.inserted_condition = Condition()

    def put(self, item: T) -> None:
        """
        Add an element to the queue

        :param item: element to add
        """
        with self.lock:
            self.queue.append(item)
        self.inserted_condition.notify()

    @rethrow_as(WouldWaitMore, Empty)
    def __get(self, timeout, item_getter) -> T:
        self.lock.acquire()
        if len(self.queue):
            # Fast path
            try:
                return item_getter(self.queue)
            finally:
                self.lock.release()
        else:
            if timeout is None:
                while True:
                    self.lock.release()
                    self.inserted_condition.wait()
                    self.lock.acquire()
                    if len(self.queue):
                        try:
                            return item_getter(self.queue)
                        finally:
                            self.lock.release()
            else:
                with measure(timeout=timeout) as measurement:
                    while not measurement.timeouted:
                        self.lock.release()
                        # raises WouldWaitMore
                        self.inserted_condition.wait(timeout=measurement.time_remaining)
                        self.lock.acquire()
                        if len(self.queue):
                            try:
                                return item_getter(self.queue)
                            finally:
                                self.lock.release()
                    else:
                        self.lock.release()
                        raise Empty('queue is empty')

    def get(self, timeout: tp.Optional[float] = None) -> T:
        """
        Get an element.

        :param timeout: maximum amount of seconds to wait. Default value of None
            means wait as long as necessary
        :return: the item
        :raise Empty: queue was empty
        """
        return self.__get(timeout, lambda queue: queue.popleft())

    def peek(self, timeout: tp.Optional[float] = None) -> T:
        """
        Get an element without removing it from the top of the queue.

        :param timeout: maximum amount of seconds to wait. Default value of None
            means wait as long as necessary
        :return: the item
        :raise WouldWaitMore: timeout has expired
        """
        return self.__get(timeout, lambda queue: queue[0])

    def qsize(self) -> int:
        """
        Return the approximate size of the queue. Note, qsize() > 0 doesn’t
        guarantee that a subsequent get() will not block.
        :return: approximate size of the queue
        """
        return len(self.queue)




import threading
import typing as tp
from threading import Thread


class ThreadCollection:
    """
    A collection of threads.

    Create like:

    >>> class MyThread(Thread):
    >>>     def __init__(self, a):
    >>>         ...
    >>> tc = ThreadCollection.from_class(MyThread, [2, 4, 5], daemon=True)
    >>> tc.start()
    >>> tc.terminate()
    >>> tc.join()

    This also implements iteration (it will return all the threads in the collection) and
    length check.
    """

    __slots__ = ('threads', )

    def __len__(self):
        return len(self.threads)

    def __iter__(self):
        return iter(self.threads)

    @classmethod
    def get_currently_running(cls, include_main_thread: bool = True) -> 'ThreadCollection':
        """
        Get all currently running threads as thread collection

        :param include_main_thread: whether to include the main thread

        :return: a thread collection representing all currently running threads
        """
        result = []
        for thread in threading.enumerate():
            # noinspection PyProtectedMember
            if not include_main_thread and isinstance(thread, threading._MainThread):
                continue
            result.append(thread)
        return ThreadCollection(result)

    @classmethod
    def from_class(cls, cls_to_use, iteratable, **kwargs) -> 'ThreadCollection':
        """
        Build a thread collection

        :param cls_to_use: class to instantiate with
        :param iteratable: an iterable with the sole argument to this class
        """
        return ThreadCollection([cls_to_use(it, **kwargs) for it in iteratable])

    @property
    def daemon(self) -> bool:
        """
        Is any of the threads a daemon?

        Also, when used a setter sets daemon attribute.
        """
        return any(thread.daemon for thread in self.threads)

    @daemon.setter
    def daemon(self, v: bool) -> None:
        for thread in self.threads:
            thread.daemon = v

    def __init__(self, threads: tp.Sequence[Thread]):
        self.threads = list(threads)

    def append(self, thread: Thread) -> None:
        """
        Alias for :meth:`~satella.coding.concurrent.ThreadCollection.add`

        :param thread: thread to add
        """
        self.add(thread)

    def add(self, thread: Thread) -> None:
        """
        Add a thread to the collection

        :param thread: thread to add
        """
        self.threads.append(thread)

    def start(self) -> None:
        """
        Start all threads
        """
        for thread in self.threads:
            thread.start()

    def terminate(self, *args, **kwargs) -> None:
        """
        Call terminate() on all threads that have this method
        """
        for thread in self.threads:
            try:
                thread.terminate(*args, **kwargs)
            except AttributeError:
                pass

    def join(self) -> None:
        """Join all threads"""
        for thread in self.threads:
            thread.join()

    def is_alive(self) -> bool:
        """
        Is at least one thread alive?
        """
        return any(thread.is_alive() for thread in self.threads)



import subprocess
import typing as tp

from satella.coding.recast_exceptions import silence_excs
from .coding.concurrent import call_in_separate_thread
from .exceptions import ProcessFailed

__all__ = ['call_and_return_stdout', 'read_nowait']

from .time import parse_time_string


@call_in_separate_thread(daemon=True)
@silence_excs((IOError, OSError))
def read_nowait(process: subprocess.Popen, output_list: tp.List[str]):
    """
    This spawns a thread to read given process' stdout and append it to a list, in
    order to prevent buffer filling up completely.

    To retrieve entire stdout after process finishes do

    >>> ''.join(list)

    This thread will terminate automatically after the process closes it's stdout or finishes.
    """
    while True:
        with silence_excs(subprocess.TimeoutExpired):
            process.wait(timeout=0.1)
            line = process.stdout.read(2048)
            if line:
                output_list.append(line)
            else:
                break


def call_and_return_stdout(args: tp.Union[str, tp.List[str]],
                           timeout: tp.Optional[tp.Union[str, int]] = None,
                           encoding: tp.Optional[str] = None,
                           expected_return_code: tp.Optional[int] = None,
                           **kwargs) -> tp.Union[bytes, str]:
    """
    Call a process and return it's stdout.

    Everything in kwargs will be passed to subprocess.Popen

    A bytes object will be returned if encoding is not defined, else stdout will be decoded
    according to specified encoding.

    :param args: arguments to run the program with. Can be either a string or a list of strings.
    :param timeout: amount of seconds to wait for the process result. If process does not complete
        within this time, it will be sent a SIGKILL. Can be also a time string. If left at default,
        ie. None, timeout won't be considered at all.
    :param encoding: encoding with which to decode stdout. If none is passed, it will be returned as
        a bytes object
    :param expected_return_code: an expected return code of this process. 0 is the default. If
        process returns anything else, ProcessFailed will be raise. If left default (None) return
        code won't be checked at all
    :raises ProcessFailed: process' result code was different from the requested
    :raises TimeoutError: timeout was specified and the process didn't complete
    """
    kwargs['stdout'] = subprocess.PIPE

    stdout_list = []

    proc = subprocess.Popen(args, **kwargs)
    fut = read_nowait(proc, stdout_list)

    if timeout is not None:
        timeout = parse_time_string(timeout)

    try:
        proc.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        proc.wait()
        raise TimeoutError('Process did not complete within %s seconds' % (timeout,))
    finally:
        fut.result()

    if encoding is None:
        result = b''.join(stdout_list)
    else:
        result = ''.join((row.decode(encoding) for row in stdout_list))

    if expected_return_code is not None:
        if proc.returncode != expected_return_code:
            raise ProcessFailed(proc.returncode, result)

    return result



from abc import ABCMeta, abstractmethod


class StrEqHashableMixin(metaclass=ABCMeta):
    """
    A mix-in that outfits your class with an __eq__ and __hash__ operator that take
    their values from __str__ representation of your class.
    """
    __slots__ = ()

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __hash__(self) -> int:
        return hash(str(self))

    @abstractmethod
    def __str__(self) -> str:
        pass


class ReprableMixin:
    """
    A sane __repr__ default.

    This takes the values for the __repr__ from repr'ing list of fields defined as
    class property _REPR_FIELDS.

    Set an optional class property of _REPR_FULL_CLASSNAME for __repr__ to output the repr
    alongside the module name.

    Example:

    >>> class Test(ReprableMixin):
    >>>     _REPR_FIELDS = ('v', )
    >>>     def __init__(self, v, **kwargs):
    >>>         self.v = v
    >>>
    >>> assert repr(Test(2)) == "Test(2)"
    >>> assert repr(Test('2') == "Test('2')")
    """
    __slots__ = ()
    _REPR_FIELDS = ()

    def __repr__(self):
        fragments = []
        if getattr(self, '_REPR_FULL_CLASSNAME', False):
            fragments = ['%s%s' % ((self.__class__.__module__ + '.')
                                   if self.__class__.__module__ != 'builtins' else '',
                                   self.__class__.__qualname__)]
        if not fragments:
            fragments = [self.__class__.__name__]
        fragments.append('(')
        arguments = []
        for field_name in self._REPR_FIELDS:
            arguments.append(repr(getattr(self, field_name)))
        fragments.append(', '.join(arguments))
        fragments.append(')')
        return ''.join(fragments)



import typing as tp
from satella.coding.typing import U, K
import bisect


def linear_interpolate(series: tp.Sequence[tp.Tuple[K, U]], t: K,
                       clip: bool = False) -> U:
    """
    Given a sorted (ascending) series of (t_value, y_value) interpolating linearly a function of
    y=f(t) compute a linear approximation of f at t
    of two closest values.

    t must be larger or equal to t_min and smaller or equal to t_max

    :param series: series of (t, y) sorted by t ascending
    :param t: t to compute the value for
    :param clip: if set to True, then values t: t<t_min f(t_min) will be returned
        and for values t: t>t_max f(t_max) will be returned
    :return: return value
    :raise ValueError: t was smaller than t_min or greater than t_max
    """
    if t < series[0][0]:
        if clip:
            return series[0][1]
        else:
            raise ValueError('t smaller than t_min')
    if t > series[-1][0]:
        if clip:
            return series[-1][1]
        else:
            raise ValueError('t greater than t_max')

    if t == series[0][0]:
        return series[0][1]

    i = bisect.bisect_left([y[0] for y in series], t) - 1

    if i == len(series)-1:
        return series[-1][1]

    t1, v1 = series[i]
    t2, v2 = series[i+1]

    assert t1 <= t <= t2, 'Series not sorted!'
    return (v2 - v1) / (t2 - t1) * (t - t1) + v1



import math
import typing as tp


# shamelessly taken from
# http://code.activestate.com/recipes/511478-finding-the-percentile-of-the-values/)
def percentile(n: tp.List[float], percent: float) -> float:
    """
    Find the percentile of a list of values.

    :param n: - is a list of values. Note this MUST BE already sorted.
    :param percent: - a float value from 0.0 to 1.0.

    :return: the percentile of the values
    """
    k = (len(n) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return n[int(k)]
    d0 = n[int(f)] * (c - k)
    d1 = n[int(c)] * (k - f)
    return d0 + d1



import collections
import itertools
import typing as tp
import warnings

from ..decorators import for_argument, wraps
from ..recast_exceptions import rethrow_as, silence_excs
from ..typing import Iteratable, T, U, Predicate, V, K


def iterate_callable(clbl: tp.Callable[[int], V], start_from: int = 0,
                     exc_classes=(IndexError, ValueError)) -> tp.Iterator[V]:
    """
    Given a callable that accepts an integer and returns the n-th entry, iterate over
    it until it starts to throw some exception.

    :param clbl: callable to call
    :param start_from: number to start from
    :param exc_classes: exceptions that being thrown show that the list was exhausted

    :return: an iterator
    """
    for i in itertools.count(start_from):
        try:
            yield clbl(i)
        except exc_classes:
            return


def try_close(iterator: tp.Iterator) -> None:
    """
    Try to invoke close() on an iterator. Do nothing if provided iterator doesn't have a .close()
    method.

    :param iterator: iterator to close
    """
    try:
        iterator.close()
    except AttributeError:
        pass


def length(iterator: Iteratable) -> int:
    """
    Return the length of an iterator, exhausting it by the way
    """
    i = 0
    for _ in iterator:
        i += 1
    return i


def f_range(*args: float) -> tp.Iterator[float]:
    """
    A range() that supports float.

    Note that this behaves correctly when given a negative step.

    Call either:

    >>> f_range(stop)   # will start from 0 and step 1
    >>> f_range(start, stop)    # will start from start and continue until the result is gte stop
    >>> # will start from start and continue by step until the result is gte stop
    >>> f_range(start, stop, step)

    :raise TypeError: invalid number of arguments
    """
    len_args = len(args)
    if len_args == 1:
        start, step = 0, 1
        stop = args[0]
    elif len_args == 2:
        start, stop = args
        step = 1
    elif len_args == 3:
        start, stop, step = args
    else:
        raise TypeError('Invalid number of arguments')

    def iterate(f_start, f_stop, f_step):
        if f_step > 0:
            while f_start < f_stop:
                yield f_start
                f_start += f_step
        else:
            while f_start > f_stop:
                yield f_start
                f_start += f_step

    return iterate(start, stop, step)


def append_sequence(seq: tp.Iterator[tuple], *elems_to_append) -> tp.Iterator[tuple]:
    """
    Return an iterator which append elem_to_append to every tuple in seq.

    Example:

    >>> a = [(1, ), (2, ), (3, )]
    >>> assert list(append_sequence(a, 1, 2)) == [(1, 1, 2), (2, 1, 2), (3, 1, 2)]

    If every element of seq is not a tuple, it will be cast to one.

    :param seq: sequence to append
    :param elems_to_append: element(s) to append
    :return: an iterator
    """
    for tpl in seq:
        if not isinstance(tpl, tuple):
            tpl = tuple(tpl)
        yield tpl + elems_to_append


def walk(obj: T, child_getter: tp.Callable[[T], tp.Optional[tp.List[T]]] = list,
         deep_first: bool = True,
         leaves_only: bool = False) -> tp.Iterator[T]:
    """
    Return every node of a nested structure.

    :param obj: structure to traverse. This will not appear in generator
    :param child_getter: a callable to return a list of children of T.
        Should return an empty list or None of there are no more children.
    :param deep_first: if True, deep first will be returned, else it will be breadth first
    :param leaves_only: if True, only leaf nodes (having no children) will be returned
    """
    a = ConstruableIterator(child_getter(obj))
    for o in a:
        children = child_getter(o)
        if children is not None:
            try:
                child_len = len(children)
            except TypeError:
                child_len = 0

            if child_len:
                if deep_first:
                    a.add_many_immediate(children)
                else:
                    a.add_many(children)
                if leaves_only:
                    continue

        yield o


class ConstruableIterator:
    """
    An iterator that you can attach arbitrary things at the end and consume them during iteration.
    Eg:

    >>> a = ConstruableIterator([1, 2, 3])
    >>> for b in a:
    >>>   if b % 2 == 0:
    >>>       a.add(6)

    All arguments you provide to the constructor will be passed to underlying deque
    """
    __slots__ = ('entries',)

    def __init__(self, *args, **kwargs):
        self.entries = collections.deque(*args, **kwargs)

    def add_immediate(self, t: T) -> None:
        """
        Schedule given value to be iterated over during the next __next__ call

        :param t: value to iterate over
        """
        self.entries.appendleft(t)

    def add_many_immediate(self, t: tp.Iterable[T]) -> None:
        """
        Schedule given values to be iterated over during the next __next__ call

        :param t: values to iterate over
        """
        for i, entry in enumerate(t):
            self.entries.insert(i, entry)

    def add(self, t: T) -> None:
        """
        Schedule given value to be iterated over after current items

        :param t: value to iterate over
        """
        self.entries.append(t)

    def add_many(self, t: tp.Iterable[T]) -> None:
        """
        Schedule given values to be iterated over after current items

        :param t: iterable of values
        """
        self.entries.extend(t)

    def __iter__(self) -> 'ConstruableIterator':
        return self

    @rethrow_as(IndexError, StopIteration)
    def __next__(self) -> T:
        return self.entries.popleft()

    def __length_hint__(self) -> int:
        return len(self.entries)


class AlreadySeen(tp.Generic[K]):
    """
    Class to filter out unique objects. Objects must be hashable, barring that
    they must be eq-able, however passing it an non-hashable object will result in
    O(n^2) complexity, as the class uses a list to keep track of the objects.

    Usage:

    >>> als = AlreadySeen()
    >>> for elem in sequence:
    >>>     if als.is_unique(elem):
    >>>         ... process the element ...
    """
    __slots__ = ('set',)

    def __init__(self):
        self.set = None     # type: tp.Union[list, set]

    def is_unique(self, key: K) -> bool:
        """
        Has the element been spotted first time?

        Add it to the set.

        :param key: element to check
        :return: whether the element was seen for the first time
        """
        try:
            hash(key)
            if self.set is None:
                self.set = set()
            if key in self.set:
                return False
            self.set.add(key)
            return True
        except TypeError:
            warnings.warn('Passed argument is not hashable, you pay with '
                          'O(n^2) complexity!', RuntimeWarning)
            if self.set is None:
                self.set = []
            if key in self.set:
                return False
            self.set.append(key)
            return True


def unique(lst: Iteratable) -> tp.Iterator[T]:
    """
    Return each element from lst, but return every element only once.

    Take care for elements of T to be __eq__-able and hashable!

    This will keep internally a set of elements encountered, and skip them if same element
    appears twice

    :param lst: iterable to process
    :return: a generator yielding unique items from lst
    """
    already_seen = set()
    for elem in lst:
        if elem not in already_seen:
            already_seen.add(elem)
            yield elem


@for_argument(iter)
def even(sq: Iteratable) -> tp.Iterator[T]:
    """
    Return only elements with even indices in this iterable (first element will be returned,
    as indices are counted from 0)
    """
    while True:
        try:
            yield next(sq)
            next(sq)
        except StopIteration:
            return


@silence_excs(StopIteration)
@for_argument(iter)
def odd(sq: Iteratable) -> tp.Iterator[T]:
    """
    Return only elements with odd indices in this iterable.
    """
    while True:
        try:
            next(sq)
            yield next(sq)
        except StopIteration:
            return


def count(sq: Iteratable, start: tp.Optional[int] = None, step: int = 1,
          start_at: tp.Optional[int] = None) -> tp.Iterator[int]:
    """
    Return a sequence of integers, for each entry in the sequence with provided step.

    Essentially the same (if step were ignored) as:

    >>> (i for i, x in enumerate(sq, start=start_at))

    .. deprecated:: 2.14.22
        Use `start` instead

    :param sq: sequence to enumerate
    :param start: alias for start_at. Prefer it in regards to start_at. Default is 0
    :param step: number to add to internal counter after each element
    :param start_at: deprecated alias for start
    :return: an iterator of subsequent integers
    """
    if start_at:
        warnings.warn('This is deprecated and will be removed in Satella 3.0. Use start instead.',
                      DeprecationWarning)
        start = start_at
    num = start or 0
    for _ in sq:
        yield num
        num += step


def iter_dict_of_list(dct: tp.Dict[T, tp.List[U]]) -> tp.Generator[tp.Tuple[T, U], None, None]:
    """
    Presents a simple way to iterate over a dictionary whose values are lists.

    This will return the dictionary key and each of the value contained in the list attached to
    the key.
    """
    for key, items in dct.items():
        for item in items:
            yield key, item


def infinite_counter(start_at: int = 0, step: int = 1) -> tp.Iterator[int]:
    """
    Infinite counter, starting at start_at

    .. deprecated:: 2.14.22
        Use `itertools.count` instead.

    :param start_at: value at which to start counting. It will be yielded as first
    :param step: step by which to progress the counter
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0, '
                  'use itertools.count() instead', DeprecationWarning)
    i = start_at
    while True:
        yield i
        i += step


def is_instance(classes: tp.Union[tp.Tuple[type, ...], type]) -> Predicate[object]:
    def inner(object_):
        return isinstance(object_, classes)

    inner.__doc__ = """Return a bool telling if object is of type %s""" % (repr(classes),)
    return inner


@for_argument(iter, iter)
def other_sequence_no_longer_than(base_sequence: Iteratable,
                                  other_sequence: Iteratable) -> tp.Iterator[T]:
    """
    Return every item in other_sequence, but limit it's p_len to that of base_sequence.

    If other_sequence is shorter than base_sequence, the shorter one will be returned.

    :param base_sequence: sequence whose p_len should be taken
    :param other_sequence: sequence to output values from
    """
    while True:
        try:
            next(base_sequence)
            yield next(other_sequence)
        except StopIteration:
            return


def shift(iterable_: tp.Union[tp.Reversible[T], Iteratable],
          shift_factor: int) -> tp.Iterator[T]:
    """
    Return this sequence, but shifted by factor elements, so that elements will appear
    sooner by factor.

    Eg:

    >>> assert list(shift([1,2, 3], 1)) == [2, 3, 1]

    However note that this will result in iterators which have negative shift to be readed entirely
    into memory (converted internally to lists). This can be avoided by passing in a Reversible
    iterable.

    :param iterable_: iterable to shift
    :param shift_factor: factor by which shift elements.
    :return: shifted sequence
    """

    if shift_factor >= 0:
        iterator = iter(iterable_)
        elements = []
        for i in range(shift_factor):
            elements.append(next(iterator))
        return itertools.chain(iterator, elements)
    else:
        if hasattr(iterable_, '__reversed__'):
            elements = take_n(reversed(iterable_), -shift_factor)
            elements = reversed(elements)
            return other_sequence_no_longer_than(iterable_, itertools.chain(elements, iterable_))
        else:
            iterator = list(iterable_)
            iterator = iterator[shift_factor:] + iterator[:shift_factor]  # shift's already negative
            return iterator


@silence_excs(StopIteration)
def zip_shifted(*args: tp.Union[Iteratable, tp.Tuple[Iteratable, int]]) -> \
        tp.Iterator[tp.Tuple[T, ...]]:
    """
    Construct an iterator, just like zip but first by cycling it's elements by it's shift factor.
    Elements will be shifted by a certain factor, this means that they will appear earlier.

    Example:

    >>> zip_shifted(([1, 2, 3, 4], 1), ([1, 2, 3, 4], 0)) == [(2, 1), (3, 2), (4, 3), (1, 4)]

    This will work on arbitrary iterators and iterables.

    Shift can be negative, in which case the last elements will appear sooner, eg.

    >>> zip_shifted(([1, 2, 3, 4], -1), ([1, 2, 3, 4], 0)) == [(4, 1), (1, 2), (2, 3), (3, 4)]

    Same memory considerations as :func:`shift` apply.

    The resulting iterator will be as long as the shortest sequence.

    .. deprecated:: 2.14.22
        Use `zip(shift(...))` instead

    :param args: a tuple with the iterator/iterable and amount of shift. If a non-tuple is given,
        it is assumed that the shift is zero.
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Use zip(shift(...)) instead!', DeprecationWarning)

    iterators = []  # type: tp.List[tp.Union[tp.Tuple[tp.Iterator[T], tp.List[T]], tp.Iterator[T]]
    for row in args:
        if not isinstance(row, tuple):
            iterators.append(row)
        else:
            iterable, shift_factor = row
            iterators.append(shift(iterable, shift_factor))
    return zip(*iterators)


@for_argument(iter)
@silence_excs(StopIteration)
def skip_first(iterator: Iteratable, n: int) -> tp.Iterator[T]:
    """
    Skip first n elements from given iterator.

    Returned iterator may be empty, if source iterator is shorter or equal to n.

    .. deprecated:: 2.14.22
        Use `itertools.islice` instead
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Please use itertools.islice instead', DeprecationWarning)

    for i in range(n):
        next(iterator)
    yield from iterator


class _ListWrapperIteratorIterator(tp.Iterator[T]):
    __slots__ = ('parent', 'pos')

    def __init__(self, parent):
        self.parent = parent
        self.pos = 0

    def __length_hint__(self) -> int:
        return len(self.parent.list)

    def __iter__(self) -> tp.Iterator[T]:
        return self

    def __next__(self) -> T:
        if self.pos >= len(self.parent.list):
            item = self.parent.next()
        else:
            item = self.parent.list[self.pos]
        self.pos += 1
        return item


class ListWrapperIterator(tp.Generic[T]):
    """
    A wrapped for an iterator, enabling using it as a normal list.

    The first time this is evaluated, list is populated with elements.

    The second time, items are taken from the list.

    It never computes more than it needs to.

    Essentially a class that lets you reuse one-shot iterators.
    """
    __slots__ = ('iterator', 'exhausted', 'list')

    def __init__(self, iterator: Iteratable):
        self.iterator = iter(iterator)
        self.exhausted = False
        self.list = []

    def exhaust(self) -> None:
        """
        Load all elements of this iterator into memory.
        """
        if self.exhausted:
            return
        for elem in self.iterator:
            self.list.append(elem)
        self.exhausted = True

    def advance_to_item(self, i: int) -> None:
        """
        Makes the list be at least i in size
        """
        if self.exhausted:
            return

        while len(self.list) < i:
            try:
                self.list.append(next(self.iterator))
            except StopIteration:
                self.exhausted = True
                return

    def __len__(self) -> int:
        self.exhaust()
        return len(self.list)

    def __getitem__(self, item: tp.Union[slice, int]) -> tp.Union[tp.List[T], T]:
        if isinstance(item, int):
            if len(self.list) < item + 1:
                self.advance_to_item(item + 1)
        else:
            self.advance_to_item(item.stop)
        return self.list[item]

    def next(self) -> T:
        """
        Get the next item

        :raises StopIteration: next element is not available due to iterator finishing
        """
        if self.exhausted:
            raise StopIteration()
        else:
            item = next(self.iterator)
            self.list.append(item)
            return item

    def __iter__(self) -> tp.Iterator[T]:
        return _ListWrapperIteratorIterator(self)


@silence_excs(StopIteration)
@for_argument(iter)
def stop_after(iterator: Iteratable, n: int) -> tp.Iterator[T]:
    """
    Stop this iterator after returning n elements, even if it's longer than that.

    The resulting iterator may be shorter than n, if the source element is so.

    .. deprecated:: 2.14.22
        Use `itertools.islice` instead

    :param iterator: iterator or iterable to examine
    :param n: elements to return
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Please use itertools.islice instead', DeprecationWarning)

    for i in range(n):
        yield next(iterator)


@for_argument(iter)
def n_th(iterator: Iteratable, n: int = 0) -> T:
    """
    Obtain n-th element (counting from 0) of an iterable

    :param iterator: iterable to process
    :param n: element to return. Note that we're counting from 0
    :raises IndexError: iterable was too short
    """
    try:
        for i in range(n):
            next(iterator)
        return next(iterator)
    except (StopIteration, GeneratorExit):
        raise IndexError('Iterable was too short')


class IteratorListAdapter:
    """
    A wrapper around an iterator that enables it to be processed as a list.

    Ie. the generator will now support __contains__, __len__ and __getitem__.
    If a call to such a method is done, the generator will be unfolded in memory so this
    might take a ton of memory! You've been warned!

    :param iterator:
    """
    __slots__ = ('unfolded', 'iterator', 'list', 'pointer')

    def __init__(self, iterator: tp.Iterator):
        self.iterator = iter(iterator)
        self.list = None
        self.unfolded = False
        self.pointer = 0

    def __iter__(self):
        if self.unfolded:
            return iter(self.list)
        else:
            return self

    @rethrow_as(IndexError, StopIteration)
    def __next__(self):
        if self.unfolded:
            self.pointer += 1
            return self.list[self.pointer - 1]  # throws: IndexError
        else:
            return next(self.iterator)

    def __unfold(self):
        if self.unfolded:
            return
        self.list = list(self.iterator)
        self.unfolded = True
        self.pointer = 0

    def __getitem__(self, item):
        self.__unfold()
        return self.list[item]

    def __contains__(self, item) -> bool:
        self.__unfold()
        return item in self.list

    def __len__(self) -> int:
        self.__unfold()
        return len(self.list)


@silence_excs(StopIteration, returns=True)
def is_empty(iterable: Iteratable, exhaust: bool = True) -> bool:
    """
    Checks whether an iterator is empty.

    This will exhaust the iterator if exhaust is left at default, or True

    :param iterable: iterator to check
    :param exhaust: if set to False, at most a single element will be consumed
        from the iterator
    :return: whether the iterator is empty
    """
    iterator = iter(iterable)
    if exhaust:
        i = 0
        for _ in iterator:
            i += 1
        return i == 0
    else:
        next(iterator)
        return False


def map_list(fun: tp.Callable, iterable: Iteratable) -> tp.List:
    """
    A syntactic sugar for

    >>> list(map(fun, iterable))

    :param fun: function to apply
    :param iterable: iterable to iterate over
    """
    return list(map(fun, iterable))


def to_iterator(fun):
    """
    Convert function to an iterator. You can replace the following code:

    >>> def iterator(x):
    >>>     for y in x:
    >>>         yield fun(y)

    with

    >>> @to_iterator
    >>> def fun(y):
    >>>     ...

    and now call fun instead of iterator. fun will accept a single argument - the iterable,
    and assume that the function you decorate also takes a single argument - the item
    """

    @wraps(fun)
    def inner(iterable):
        for item in iterable:
            yield fun(item)

    return inner


def smart_zip(*iterators: Iteratable) -> tp.Iterator[tp.Tuple[T, ...]]:
    """
    Zip in such a way that resulted tuples are automatically expanded.

    Ie:

    >>> b = list(smart_zip([(1, 1), (1, 2)], [1, 2]))
    >>> assert b == [(1, 1, 1), (1, 2, 2)]

    Note that an element of the zipped iterator must be a tuple (ie. isinstance tuple)
    in order for it to be appended to resulting iterator element!

    :param iterators: list of iterators to zip together
    :return: an iterator zipping the arguments in a smart way
    """
    for row in zip(*iterators):
        a = []
        for elem in row:
            if isinstance(elem, tuple):
                a.extend(elem)
            else:
                a.append(elem)
        yield tuple(a)


def enumerate2(iterable: Iteratable, start: int = 0,
               step: int = 1) -> tp.Iterator[tp.Tuple[int, T]]:
    """
    Enumerate with a custom step

    :param iterable: iterable to enumerate
    :param start: value to start at
    :param step: step to add during each iteration
    """
    v = start
    for item in iterable:
        yield v, item
        v += step


def smart_enumerate(iterator: Iteratable, start: int = 0,
                    step: int = 1) -> tp.Iterator[tp.Tuple]:
    """
    An enumerate that talks pretty with lists of tuples. Consider

    >>> a = [(1, 2), (3, 4), (5, 6)]
    >>> for i, b in enumerate(a):
    >>>     c, d = b
    >>>     ...

    This function allows you just to write:
    >>> for i, c, d in enumerate(a):
    >>>     ...

    Note that elements in your iterable must be either a list of a tuple for that to work,
    or need to be able to be coerced to a tuple. Otherwise, TypeError will be thrown.

    :param iterator: iterator to enumerate
    :param start: value to start counting at
    :param step: step to advance the enumeration with
    :raise TypeError: could not coerce the elements in your iterable to a tuple
    """
    i = start
    for row in iterator:
        if isinstance(row, tuple):
            yield (i,) + row
        else:
            yield (i,) + tuple(row)
        i += step


@for_argument(iter)
def take_n(iterator: Iteratable, n: int, skip: int = 0) -> tp.List[T]:
    """
    Take (first) n elements of an iterator, or the entire iterator, whichever comes first

    :param iterator: iterator to take from
    :param n: amount of elements to take
    :param skip: elements from the start to skip
    :return: list of p_len n (or shorter)
    """
    for i in range(skip):
        next(iterator)

    output = []
    for i in range(n):
        try:
            output.append(next(iterator))
        except StopIteration:
            return output
    return output



import collections
import typing as tp

from .base import EmbeddedSubmetrics, MetricLevel
from .registry import register_metric
from ..data import MetricDataCollection, MetricData


@register_metric
class LinkfailMetric(EmbeddedSubmetrics):
    """
    Metric that measures whether given link is operable.

    :param consecutive_failures_to_offline: consecutive failures needed for link to become offline
    :param consecutive_successes_to_online: consecutive successes needed for link to become online
        after a failure
    :param callback_on_online: callback that accepts an address of a link that becomes online
        and labels
    :param callback_on_offline: callback that accepts an address of a link that becomes offline
        and labels
    """
    __slots__ = ('working', 'consecutive_failures', 'consecutive_successes',
                 'callback_on_online', 'callback_on_offline',
                 'consecutive_failures_to_offline', 'consecutive_successes_to_online')

    CLASS_NAME = 'linkfail'

    def __init__(self, name: str, root_metric: 'Metric' = None,
                 metric_level: tp.Optional[MetricLevel] = None,
                 labels: tp.Optional[dict] = None, internal: bool = False,
                 consecutive_failures_to_offline: int = 100,
                 consecutive_successes_to_online: int = 10,
                 callback_on_online: tp.Callable[[int, dict], None] = lambda a, b: None,
                 callback_on_offline: tp.Callable[[int, dict], None] = lambda a, b: None,
                 *args, **kwargs):
        super().__init__(name, root_metric, metric_level, labels, internal, *args,
                         consecutive_failures_to_offline=consecutive_failures_to_offline,
                         consecutive_successes_to_online=consecutive_successes_to_online,
                         callback_on_offline=callback_on_offline,
                         callback_on_online=callback_on_online,
                         **kwargs)
        self.working = collections.defaultdict(lambda: True)  # type: tp.Dict[int, bool]
        self.consecutive_failures = collections.defaultdict(lambda: 0)  # type: tp.Dict[int, int]
        self.consecutive_successes = collections.defaultdict(lambda: 0)  # type: tp.Dict[int, int]
        self.callback_on_online = callback_on_online  # type: tp.Callable
        self.callback_on_offline = callback_on_offline  # type: tp.Callable
        self.consecutive_failures_to_offline = consecutive_failures_to_offline  # type: int
        self.consecutive_successes_to_online = consecutive_successes_to_online  # type: int

    def _handle(self, success: bool, address: int = 0, *args, **labels):
        if self.embedded_submetrics_enabled or labels:
            return super()._handle(success, address=address, *args, **labels)
        if success:
            self.consecutive_failures[address] = 0
            self.consecutive_successes[address] += 1
            if not self.working[address]:
                if self.consecutive_successes[address] == self.consecutive_successes_to_online:
                    self.working[address] = True
                    self.callback_on_online(address, self.labels)
        else:
            self.consecutive_failures[address] += 1
            self.consecutive_successes[address] = 0

            if self.working[address]:
                if self.consecutive_failures[address] == self.consecutive_failures_to_offline:
                    self.working[address] = False
                    self.callback_on_offline(address, self.labels)

    def to_metric_data(self) -> MetricDataCollection:
        mdc = MetricDataCollection()
        keys = set(self.consecutive_successes.keys())
        keys = keys.union(set(self.consecutive_failures.keys()))
        for address in keys:
            labels = self.labels.copy()
            if keys != {0}:
                labels.update(address=address)
            mdc += MetricData(self.name + '.consecutive_failures',
                              self.consecutive_failures[address],
                              labels, self.get_timestamp(), self.internal)
            mdc += MetricData(self.name + '.consecutive_successes',
                              self.consecutive_successes[address],
                              labels, self.get_timestamp(), self.internal)
            mdc += MetricData(self.name + '.status', int(self.working[address]),
                              self.labels, self.get_timestamp(), self.internal)
        return mdc
